PAGINATION_MODEL = {
    "list_permission_sets": {
        "input_token": "next_token",
        "limit_key": "max_results",
        "limit_default": 100,
        "result_key": "PermissionSets",
        "unique_attribute": "permission_set_arn",
    },
}
